import re
from typing import List, Tuple, Union, Optional, Dict
from urllib.parse import urlparse, urlunparse, urljoin, ParseResult

from selectolax.parser import HTMLParser
from selectolax.parser import Node

from .utilities import _URL, _CssSelector, make_absolute, fuzzy_search
from .utilities import ilt_elements


class Ahref:
    """Basic class for extracted Ahrefs

    :param href_node: selectolax node object
    :base_url: the url of the webpage
    """

    def __init__(self, href_node: Node, base_url: _URL) -> None:
        self.href: str = href_node.attributes.get("href")
        if self.href.startswith(("#", "mailto:", "javascript:", "tel:")):
            raise ValueError("Not a normal url")

        self.base_url: str = base_url
        self.absolute_url: str = make_absolute(self.href, self.base_url)
        self.text: str = href_node.text()
        self.parsed = urlparse(self.absolute_url)
        self.absolute_url: str = make_absolute(self.href, self.base_url)
        self.is_internal: bool = self.parsed.netloc == urlparse(self.base_url).netloc
        self.attributes: Dict[str, str] = href_node.attributes

    def __repr__(self) -> str:
        return f"Ahref: {self.absolute_url}"

    def __str__(self) -> str:
        return self.absolute_url

    @property
    def path_split(self) -> List[str]:
        parsed = urlparse(self.absolute_url)

        return [item.lower() for item in parsed.path.split("/") if len(item) > 0]


class BaseParser:
    """The Basic HTML parser

    :param html: String representation of the HTML document
    :param url: URL of the page rendered

    """

    next_symbol = ["volgende", "next", "meer", "more", "ouder", "older"]
    prev_symbol = ["vorige", "previous", "nieuwe", "new"]

    def __init__(self, html: str, url: str):
        self.url: str = url
        self.parsed_url = urlparse(self.url)
        self.scheme_domain: str = (
            f"{self.parsed_url.scheme}://{self.parsed_url.hostname}/"
        )
        self._html_input: str = html
        self._tree = HTMLParser(self._html_input)
        self.links: List[Ahref] = self._get_links()

    def __repr(self) -> str:
        return f"Webpage :{self.url}"

    def css(self, selector: _CssSelector) -> List[Node]:
        return self._tree.css(selector)

    def css_first(self, selector: _CssSelector) -> Node:
        return self._tree.css_first(selector)

    def find_in_text(
        self,
        selector: _CssSelector,
        to_search: Union[List[str], str],
    ) -> List[Node]:
        """Find string in text of the Nodes."""
        # If str, make a list
        if isinstance(to_search, str):
            to_search = [to_search]

        nodes = self.css(selector)
        return [
            node
            for node in nodes
            if any(substring in node.text(deep=False) for substring in to_search)
        ]

    def find_in_text_regex(self, selector: _CssSelector, reg_str: str) -> List[Node]:
        """Find regex in text and returns the Nodes."""
        nodes = self.css(selector)
        return [node for node in nodes if re.search(reg_str, node.text(deep=False))]

    def find_in_ahref(
        self,
        to_search: Union[List[str], str],
        in_path: bool = True,
        in_text: bool = True,
        fuzzy_score: Optional[int] = None,
        internal: bool = True,
        external: bool = True,
    ) -> List[Node]:
        """Find text in the link_list. Href starting with #:, javascript:,
        mailto:, tel: are left out.

        If fuzzy int provided, returned list is ordered by fuzzy score.

        When searchin in paths, don't use '/' backslashes, just the text
        :param to_search: the string(s) to be found
        :param in_path: searches in the path of href
        "param in_text: searches in the text
        "param fuzzy: integer between 0 and 100 used for minimum fuzzy search
        "param internal: only search in internal links
        """
        if isinstance(to_search, str):
            to_search = [to_search]

        # Select list of Ahrefs from which to pick
        link_list = []
        if internal:
            link_list.extend(self.internal_links)
        if external:
            link_list.extend(self.external_links)

        # Function that creates the list to be searched
        def path_text(link: Ahref) -> List[str]:
            return_list = []
            if in_path:
                return_list.extend(link.path_split)
            if in_text:
                return_list.append(link.text)
            return [path_text.lower() for path_text in return_list]

        if fuzzy_score:
            return_list = []
            for link in link_list:
                fuzz_extract = fuzzy_search(to_search, path_text(link))
                if fuzz_extract.score >= fuzzy_score:
                    return_list.append(link)
            return return_list

        return [
            link
            for link in link_list
            if any(string.lower() in path_text(link) for string in to_search)
        ]

    @classmethod
    def get_parent(cls, node: Node, depth: int = 1) -> Node:
        parent_node = node.parent
        depth -= 1
        if not parent_node:
            return node

        if depth == 0:
            return node.parent
        else:
            return cls.get_parent(parent_node, depth)

        parent = cls.get_parent(node, depth=depth)
        return parent

    @classmethod
    def get_children(cls, node: Node):
        return [i for i in node.iter()]

    @classmethod
    def get_siblings(cls, node: Node):
        pass

    @classmethod
    def get_first_block(cls, node: Node) -> Optional[Node]:
        """returns the first ancestor that is block level,
        if node is block level, it is returned. If Node has
        no parent, None is returned
        """
        if node.tag not in ilt_elements:
            return node

        return cls.get_first_block(node.parent)

    @property
    def internal_links(self) -> List[Ahref]:
        return [link for link in self.links if link.is_internal]

    @property
    def external_links(self) -> List[Ahref]:
        return [link for link in self.links if not link.is_internal]

    @property
    def text(self) -> str:
        new_tree = HTMLParser(self._html_input)
        tags = ["head", "script", "noscript", "style", "iframe", "noembed", "noframes"]
        new_tree.strip_tags(tags)
        return new_tree.text(separator=" ").strip()

    @property
    def next_page_url(self) -> _URL:
        if (next_node := self.css_first("link[rel='next'][href]")) :
            return next_node.attributes["href"]

        # find links with any next_symbol in path or text and check in "nav"
        # or "next" in class or id
        # TODO: check parent class and id
        candidates = self.find_in_ahref(self.next_symbol, external=False)
        for candidate in candidates:
            candidate_id_class = candidate.attributes.get(
                "class", ""
            ) + candidate.attributes.get("id", "")
            if any(i in candidate_id_class for i in ["nav", "next"]):
                return candidate.absolute_url

    @property
    def previous_page_url(self) -> _URL:
        if (prev_node := self.css_first("link[rel='prev'][href]")) :
            return prev_node.attributes["href"]

        # find links with any prev_symbol in path or text and check if "nav"
        # or "next" in class or id
        # TODO: check parent class and id
        candidates = self.find_in_ahref(self.prev_symbol, external=False)
        for candidate in candidates:
            candidate_id_class_text = (
                candidate.attributes.get("class", "")
                + " "
                + candidate.attributes.get("id", "")
                + " "
                + candidate.text
            )
            if any(["nav", "next"] in candidate_id_class):
                return candidate.absolute_url

    def _get_links(self) -> List[Ahref]:
        return_list = []
        ahrefs = self._tree.css("a[href]")

        for a_node in ahrefs:
            if not a_node.attributes["href"].startswith(
                ("#", "javascript:", "mailto:", "tel:")
            ):
                return_list.append(Ahref(a_node, self.url))

        return return_list



from typing import Union, List
from urllib.parse import urlparse, urlunparse, urljoin
from dataclasses import dataclass

from fuzzywuzzy import fuzz


# Typing
_URL = str
_CssSelector = str


@dataclass
class FuzzyExtract:
    extract: str
    score: int


def make_absolute(link: str, base_url: str) -> _URL:
    """Makes a given link absolute."""

    # Parse the link with stdlib.
    parsed = urlparse(link)._asdict()

    # If link is relative, then join it with base_url.
    if not parsed["netloc"]:
        return urljoin(base_url, link)

    # Link is absolute; if it lacks a scheme, add one from base_url.
    if not parsed["scheme"]:
        parsed["scheme"] = urlparse(base_url).scheme

        # Reconstruct the URL to incorporate the new scheme.
        parsed = (v for v in parsed.values())
        return urlunparse(parsed)

    # Link is absolute and complete with scheme; nothing to be done here.
    return link


def fuzzy_search(
    to_match: Union[List[str], str], search_list: List[str]
) -> FuzzyExtract:
    if isinstance(to_match, str):
        to_match = [to_match]
    highest_match: int = 0
    extract: str = ""
    for i in search_list:
        for string in to_match:
            score = fuzz.ratio(string, i)
            if score > highest_match:
                highest_match = score
                extract = i
    return FuzzyExtract(extract, highest_match)


# List of tags that are mostly used for inline html elements
inline_elements = [
    "a",
    "abbr",
    "acronym",
    "b",
    "bdo",
    "big",
    "br",
    "button",
    "cite",
    "code",
    "del",
    "dfn",
    "em",
    "i",
    "img",
    "ins",
    "input",
    "kdb",
    "label",
    "map",
    "mark",
    "object",
    "output",
    "q",
    "samp",
    "script",
    "select",
    "small",
    "span",
    "strong",
    "sub",
    "sup",
    "textarea",
    "time",
    "tt",
    "var",
]


# List of list element tags
list_elements = [
    "li",
    "dl",
    "dt",
    "dd",
]

# List of table elements (not main table tag)
table_elements = [
    "tr",
    "th",
    "td",
]


ilt_elements = inline_elements + list_elements + table_elements



import re
import os

import pytest
import pytest_check as check
from selectolax.parser import HTMLParser

from fastparse.utilities import make_absolute
from fastparse.base_parser import Ahref
from fastparse.base_parser import BaseParser

TEST_HTML_1 = """<html>
    <head>
        <title>Een test pagina</title>
    </head>
    <body>
        <script>Een script tag</script>
        <style>Een style tag</style>
        <p class="select">Een test tag</p>
    </body>
    </html>"""

SCRIPT_PATH = os.path.realpath(__file__)
DIRR = os.path.dirname(SCRIPT_PATH)
HTML_FILE = "python.html"
with open(os.path.join(DIRR, HTML_FILE), "r") as html_file:
    TEST_HTML_PAGE = html_file.read()
TEST_HTML_URL = "https://www.python.org/"


def test_make_absolute():
    url = "/testers"
    base_url = "https://www.getevents.nl/asd/"
    absolute = make_absolute(url, base_url)
    assert absolute == "https://www.getevents.nl/testers"


def test_make_absolute_external():
    url = "//www.google.com/tester"
    base_url = "https://www.getevents.nl/tester"
    absolute = make_absolute(url, base_url)
    assert absolute == "https://www.google.com/tester"


def test_ahref_split():
    tree = HTMLParser(
        '<html><a href="https://www.google.com/tester/nog1/enq/">Google tester</a></html>'
    )
    href_node = tree.css_first("a")
    base = "https://www.google.com/"

    ahref = Ahref(href_node, base)
    check.equal(len(ahref.path_split), 3)
    check.equal(ahref.parsed.netloc, "www.google.com")
    check.equal(str(ahref), "https://www.google.com/tester/nog1/enq/")
    check.equal(ahref.parsed.scheme, "https")
    check.is_true(ahref.is_internal)


def test_relative_url():
    tree = HTMLParser('<html><a href="/test/relative">Testing</a></html>')
    href_node = tree.css_first("a")
    base = "https://www.google.com/tester1"

    ahref = Ahref(href_node, base)
    assert ahref.absolute_url == "https://www.google.com/test/relative"


def test_external_url():
    tree = HTMLParser(
        '<html><a href="https://www.getevents.nl/tester">Getevents</a></html>'
    )
    href_node = tree.css_first("a")
    base = "https://www.google.com/getevents"

    ahref = Ahref(href_node, base)
    assert not ahref.is_internal


def test_base_url():
    url = (
        "https://www.google.com/search?source=hp&ei=gF1nX5HVE5GxkwWNkaWgAw&q=url+tester"
        "&oq=url+tester&gs_lcp=CgZwc3ktYWIQAzICCAAyAggAMgIIADIGCAAQFhAeMgYIABAWEB4yBggAE"
        "BYQHjIGCAAQFhAeMgYIABAWEB4yBggAEBYQHjIGCAAQFhAeOgsILhCxAxCDARCTAjoICAAQsQMQgwE6B"
        "QgAELEDOgUILhCxAzoICC4QsQMQgwE6AgguUJwKWMMVYL8WaAFwAHgAgAGAAYgBswaSAQM2LjOYAQCg"
        "AQGqAQdnd3Mtd2l6sAEA&sclient=psy-ab&ved=0ahUKEwiRld3T7_frAhWR2KQKHY1ICTQQ4dUDCA"
        "c&uact=5"
    )
    html_string = "<p>Een stukje html</p>"

    page = BaseParser(html_string, url)
    assert page.scheme_domain == "https://www.google.com/"


def test_text():
    url = "https://www.google.com/"
    html_string = """<html>
    <head>
        <title>Een test pagina</title>
        <!--
        - DEze
        - Multiline comment moet eruit
        -->
    </head>
    <body>
        <!-- Deze comment moet eruit -->
        <script>Een script tag</script>
        <style>Een style tag</style>
        <p>Deze content moet bewaard blijven
        </p>
    </body>
    </html>"""

    page = BaseParser(html_string, url)
    assert page.text == "Deze content moet bewaard blijven"


def test_css():
    url = "https://www.getevents.nl"
    html_string = """<html>
    <head>
        <title>Een test pagina</title>
    </head>
    <body>
        <script>Een script tag</script>
        <style>Een style tag</style>
        <p class="select">Een stukje content</p>
    </body>
    </html>"""

    page = BaseParser(html_string, url)
    nodes = page.css("p.select")
    assert nodes[0].text() == "Een stukje content"


def test_css_first():
    url = "https://www.getevents.nl"
    html_string = """<html>
    <head>
        <title>Een test pagina</title>
    </head>
    <body>
        <script>Een script tag</script>
        <style>Een style tag</style>
        <p class="select">Een stukje content</p>
    </body>
    </html>"""

    page = BaseParser(html_string, url)
    node = page.css_first("p.select")
    assert node.text() == "Een stukje content"


def test_find_in_text():
    url = "https://www.google.com"
    html_string = """<html>
    <head>
        <title>Een test pagina</title>
    </head>
    <body>
        <script>Een script tag</script>
        <style>Een style tag</style>
        <p class="select">Een test tag</p>
    </body>
    </html>"""

    page = BaseParser(html_string, url)
    nodes_p = page.find_in_text("p", "test")
    nodes_star = page.find_in_text("*", "test")
    check.equal(nodes_p[0].html, '<p class="select">Een test tag</p>')
    check.equal(len(nodes_star), 2)


# TODO add test_find_in_text_fuzzy


def test_find_in_text_regex():
    url = "https://www.google.com/"
    page = BaseParser(TEST_HTML_1, url)
    reg = re.compile(r"t\wst")
    nodes = page.find_in_text_regex("p", reg)
    assert nodes[0].text() == "Een test tag"


def test_links():
    page = BaseParser(TEST_HTML_PAGE, TEST_HTML_URL)
    assert len(page.links) == 197


def test_internal_links():
    page = BaseParser(TEST_HTML_PAGE, TEST_HTML_URL)
    nr_absolute_links = len(page.internal_links)
    domain_in_links = len(
        [i for i in page.internal_links if TEST_HTML_URL in i.absolute_url]
    )
    check.equal(nr_absolute_links, domain_in_links)
    check.is_true(page.internal_links)


def test_external_links():
    page = BaseParser(TEST_HTML_PAGE, TEST_HTML_URL)
    check.is_not_in(TEST_HTML_PAGE, [i.absolute_url for i in page.external_links])
    check.is_true(page.external_links)


def test_find_in_ahref():
    page = BaseParser(TEST_HTML_PAGE, TEST_HTML_URL)
    search_string = "learn more"
    first_found = page.find_in_ahref(search_string)[0]
    assert first_found.href == "/doc/"


def test_find_in_ahref_internal():
    page = BaseParser(TEST_HTML_PAGE, TEST_HTML_URL)
    search_string = "Python 3.8.6 is now available"
    internal_links = page.find_in_ahref(search_string, external=False)
    all_links = page.find_in_ahref(search_string)
    check.is_true(all_links)
    check.equal(len(internal_links), 0)


def test_find_in_ahref_internal_searchlist():
    page = BaseParser(TEST_HTML_PAGE, TEST_HTML_URL)
    search_list = ["donate", "facebook"]
    links = page.find_in_ahref(search_list)
    print(links)
    assert len(links) == 2


def test_find_in_ahref_path():
    page = BaseParser(TEST_HTML_PAGE, TEST_HTML_URL)
    search_string = "doc"
    links = page.find_in_ahref(search_string, in_text=False, internal=True)
    assert links[0].text.lower() == "documentation"


def test_find_in_ahref_text():
    page = BaseParser(TEST_HTML_PAGE, TEST_HTML_URL)
    search_string = "documentation"
    links = page.find_in_ahref(search_string, in_path=False)
    assert links[0].href == "/doc/"


def test_find_in_ahref_fuzzy():
    page = BaseParser(TEST_HTML_PAGE, TEST_HTML_URL)
    search_string = "donate"
    links = page.find_in_ahref(search_string, fuzzy_score=90)
    print(links)
    assert (
        str(links[0])
        == "https://psfmember.org/civicrm/contribute/transact?reset=1&id=2"
    )


def test_next_page_rel_next():
    html_string = """<html>
    <head>
        <title>Page 1</title>
        <link rel="next" href="https://www.getevents.nl/p2" />
    </head>
    <body>
        <p>Een p tag</p>
    </body>
    </html>"""
    mock_url = "https://www.getevents.nl/"
    page = BaseParser(html_string, mock_url)
    assert page.next_page_url == "https://www.getevents.nl/p2"


def test_prev_page_rel_prev():
    html_string = """<html>
    <head>
        <title>Page 2</title>
        <link rel="prev" href="https://www.getevents.nl/"
    </head>
    <body>
        <p>Een p tag</p>
    </body>
    </html>"""
    mock_url = "https://www.getevents.nl/p2"
    page = BaseParser(html_string, mock_url)
    assert page.previous_page_url == "https://www.getevents.nl/"


def test_next_page_url():
    html_string = """<html>
    <head>
        <title>Page 1</title>
    </head>
    <body>
        <p>Een p tag</p>
        <a href="/p2" class="next">Next</a>
    </body>
    </html>"""
    mock_url = "https://www.getevents.nl/"
    page = BaseParser(html_string, mock_url)
    assert page.next_page_url == "https://www.getevents.nl/p2"


def test_get_parent():
    html_string = """<html>
    <head>
        <title>Test page</title>
    </head>
    <body class="grand_parent">
        <article class="parent">
            <h1>Headline</h1>
            <p>Some text ...</p>
        </article>
        <sidebar>
            <div>
                Widget
            </div>
        </sidebar>
    </body>
    </html>"""
    mock_url = "https://www.getevents.nl/"
    page = BaseParser(html_string, mock_url)
    p_node = page.css_first("p")
    parent = page.get_parent(p_node)
    grand_parent = page.get_parent(p_node, 2)
    check.equal(parent.attributes["class"], "parent")
    check.equal(grand_parent.attributes["class"], "grand_parent")


def test_get_children():
    html_string = """<html>
    <head>
        <title>Test page</title>
    </head>
    <body class="grand_parent">
        <article class="parent">
            <h1>Headline</h1>
            <p>Some text ...</p>
            <p>Antother tag</p>
            <button>Click me!</button>
        </article>
        <sidebar>
            <div>
                Widget
            </div>
        </sidebar>
    </body>
    </html>"""
    mock_url = "https://www.getevents.nl/"
    page = BaseParser(html_string, mock_url)
    a_node = page.css_first("article")
    children = BaseParser.get_children(a_node)
    assert len(children) == 4


def test_get_first_block():
    html_string = """<html>
    <head>
        <title>Test Page</title>
    </head>
    <body>
        <article>
            <ul class="blockParent">
                <li>test1</li>
                <li id="selectMe">test2</li>
                <li>test3</li>
            </ul>
        </article>
    </body>
    </html>"""
    mock_url = "https://www.getevents.nl/"
    page = BaseParser(html_string, mock_url)
    li_node = page.css_first("li#selectMe")
    block_parent = page.get_first_block(li_node)
    assert block_parent.attributes["class"] == "blockParent"


def test_get_first_block_with_inner_inline():
    html_string = """<html>
    <head>
        <title>Test Page</title>
    </head>
    <body>
        <atricle>
            <ul class="BlockParent">
                <li>Test 1</li>
                <li>Test 2</li>
                <li>Test <strong id="SelectMe">Nog 1</strong></li>
            </ul>
        </article>
    </body>
    </html>"""
    mock_url = "https://www.getevents.nl/"
    page = BaseParser(html_string, mock_url)
    strong_node = page.css_first("strong#SelectMe")
    block_parent = page.get_first_block(strong_node)
    assert block_parent.attributes["class"] == "BlockParent"




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="fastparser",
    version="0.1",
    author="Erik Meijer",
    author_email="erik@thewildweb.nl",
    description="A HTML Parser with some handy Utilities",
    long_description=long_description,
    url="https://github.com/Thewildweb/fastparser",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Development Status :: 3 - Alpha",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires=">=3.6",
)
